# NaaF: AWS Lambda

NiFi as a Function - AWS Lambda

This module provides the necessary components for running NiFi as a Function (NiFi Stateless) on AWS Lambda.


## Creating Lambda Function

To create a Lambda Function that is able to run NiFi as a Function, follow these steps.
- Navigate to the AWS Lambda page (for example, https://us-west-2.console.aws.amazon.com/lambda/home)
- Click the "Create function" button on the top-right of the page.
- Choose "Author from scratch" (the default)
- Give the function a name. For the Runtime, choose "Java 8 on Amazon Linux 2". Use the default architecture, x86_64
- Optionally, configure the VPN and the Execution Role.
- Click "Create function"

Once the function has been created, we need to provide the code for the function to run. In the "Code source" tab, click the button labeled "Upload from."
Choose ".zip or .jar file" and upload the .zip file that was created from compiling the `naaf-aws-assembly` module.
Alternatively, you may choose "Upload from Amazon S3 location" and enter the Amazon S3 URL for the .zip file.

Update the Runtime Settings by clicking the "Edit" button. The "Handler" should be set to `com.cloudera.naaf.aws.lambda.StatelessNiFiFunctionHandler::handleRequest`.
Click Save.


## Configuration

### General configuration

In the Configuration screen, Lambda provides a "General configuration" tab. This supplies two very important configuration elements: the amount of memory
allocated to the Lambda and the Timeout. The appropriate values for these two settings will depend on the dataflow to be deployed. Typically, it is recommended
to start with a Memory allocation of 1536-2048 MB (1.5 - 2 GB). This amount of memory may not be necessary, or more may be required, and this can be refined later.
While NiFi as a Function tends to perform very well during a "Warm Start," a Cold Start that must source extensions may take several seconds to initialized. As a result,
it is recommended to set the Timeout to at least 30 seconds. If the dataflow to run reaches out to many other services or performs complex computation, it may be necessary
to use a larger value, even several minutes.
 

### Configuring Environment Variables

The function must now be configured to specify the dataflow to run and any necessary runtime configuration.
Configuration is provided using Environment Variables. Navigate to the "Configuration" tab and choose "Environment variables" in the left-hand menu.

Click the "Edit" button on the Environment variables pane.

The UI provides the user the ability to add one or more environment variables in order to configure the function. 
The following Environment Variables are supported.

| Variable Name                 | Description                                                                                                                                                                                                                                                                                                                                                                                                                                                                               | Required | Default Value |
|-------------------------------|-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------| -------- | ------------- |
| FLOW_CRN                      | The Cloudera Resource Name (CRN) for the dataflow that is to be run. The dataflow must be stored in the DataFlow Service's Catalog. This CRN should indicate the specific version of the dataflow and as such will end with some suffix such as `/v.1`                                                                                                                                                                                                                                    | true | -- |
| DF_PRIVATE_KEY                | The Private Key to use for accessing Cloudera's DataFlow service. The Private Key and Access Key are used to authenticate with the DataFlow Service and must provide the necessary authorizations to access the specified dataflow.                                                                                                                                                                                                                                                       | true | -- |
| DF_ACCESS_KEY                 | The Access Key to use for accessing Cloudera's DataFlow service. The Private Key and Access Key are used to authenticate with the DataFlow Service and must provide the necessary authorizations to access the specified dataflow.                                                                                                                                                                                                                                                        | true | -- |
| INPUT_PORT                    | The name of the Input Port to use, if the dataflow has at least one.  If the dataflow has no Input Ports, this variable must not be specified.  If the specified dataflow has more than one Input Port at the root group level, this environment variable must be specified, indicating the name of the Input Port to queue up the AWS Lambda notification. If only one Input Port exists, this variable is not required but if specified must properly match the name of the Input Port. | false | -- |
| OUTPUT_PORT                   | The name of the Output Port to retrieve the results from. If no Output Port exists, this does not need to be specified and no data will be returned. If at least one Output Port exists in the dataflow, this variable can be used to determine the name of the Output Port whose data will be sent along as output of the function. Otherwise, see the section below to understand how the appropriate Output Port is determined.                                                        | false | -- |
| FAILURE_PORTS                 | A comma-separated list of Output Ports that exist at the root group level of the dataflow. If any FlowFile is sent to one of these Output Ports, the function invocation is considered a failure. See section below on Output Ports for more information.                                                                                                                                                                                                                                 | false | -- |
| DEFAULT_PARAM_CONTEXT         | If the dataflow makes use of a Parameter Context, the Lambda function will look for a Secret in the AWS Secrets Manager with the same name. If no Secret can be found with that name, this specifies the name of the Secret to default to. See section below on Specifying Parameters for more information.                                                                                                                                                                               | false | -- |
| PARAM_CONTEXT_*               | If the dataflow makes use of Parameter Contexts, this environment variable provides a mechanism for mapping a Parameter Context to one or more alternate Secrets in the AWS Secrets Manager, in a comma-separated list. See the section below on Specifying Parameters for more information.                                                                                                                                                                                              | false | -- |
| CONTENT_REPO                  | The contents of the FlowFiles can be stored either in memory, on the JVM heap, or on disk. If this environment variable is specified, it specifies the path to a directory where the content should be stored. If not specified, the content is held in memory.                                                                                                                                                                                                                           | false | -- |
| DF_SERVICE_URL                | The Base URL for the Cloudera Dataflow Service.                                                                                                                                                                                                                                                                                                                                                                                                                                           | false | https://api.us-west-1.cdp.cloudera.com |
| NEXUS_URL                     | The Base URL for a Nexus Repository to use for downloading any NiFi Archives (NARs) needed for running the dataflow.                                                                                                                                                                                                                                                                                                                                                                      | false | https://repository.cloudera.com/artifactory/cloudera-repos/ |
| STORAGE_BUCKET                | An S3 bucket in which to look in for custom extensions / NiFi Archives (NARs) and resources. See the section below on [S3 Bucket Storage](#s3-bucket-storage) for more information.                                                                                                                                                                                                                                                                                                       | false | -- |
| STORAGE_EXTENSIONS_DIRECTORY  | The directory in the S3 bucket to look in for custom extensions / NiFi Archives (NARs). See the section below on [S3 Bucket Storage](#s3-bucket-storage) for more information.                                                                                                                                                                                                                                                                                                            | false | extensions |
| STORAGE_RESOURCES_DIRECTORY   | The directory in the S3 bucket to look in for custom resources. See the section below on [S3 Bucket Storage](#s3-bucket-storage) for more information.                                                                                                                                                                                                                                                                                                                                    | false | resources |
| WORKING_DIR                   | The working directory, where NAR files will be expanded.                                                                                                                                                                                                                                                                                                                                                                                                                                  | false | /tmp/working |
| EXTENSIONS_DOWNLOAD_DIR       | The directory to which missing extensions / NiFi Archives (NARs) will be downloaded. See the section below on Providing Custom Extensions / NARs for more information.                                                                                                                                                                                                                                                                                                                    | false | /tmp/extensions |
| EXTENSIONS_DIR_*              | Specifies read-only directories that may contain custom extensions / NiFi Archives (NARs). See the section below on Providing Custom Extensions / NARs for more information.                                                                                                                                                                                                                                                                                                              | false | -- |
| DISABLE_STATE_PROVIDER        | If true, disables the DynamoDB dataflow state provider, even if the dataflow has stateful processors.                                                                                                                                                                                                                                                                                                                                                                                     | false | -- |
| DYNAMODB_STATE_TABLE          | The DynamoDB table name where the state will be stored                                                                                                                                                                                                                                                                                                                                                                                                                                    | false | `nifi_state` |
| DYNAMODB_BILLING_MODE         | Sets the DynamoDB Billing Mode (`PROVISIONED` or `PAY_PER_REQUEST`)                                                                                                                                                                                                                                                                                                                                                                                                                       | false | PAY_PER_REQUEST |
| DYNAMODB_WRITE_CAPACITY_UNITS | Sets the DynamoDB Write Capacity Units, when using `PROVISIONED` Billing Mode                                                                                                                                                                                                                                                                                                                                                                                                             | true if using `PROVISIONED` Billing Mode | -- |
| DYNAMODB_READ_CAPACITY_UNITS  | Sets the DynamoDB Read Capacity Units, when using `PROVISIONED` Billing Mode                                                                                                                                                                                                                                                                                                                                                                                                              | true if using `PROVISIONED` Billing Mode | -- |
| KRB5_FILE                     | Specifies the filename of the krb5.conf file. This is necessary only if connecting to a Kerberos-protected endpoint. See section below on configuring Kerberos.                                                                                                                                                                                                                                                                                                                           | false | `/etc/krb5.conf` |



## Output Ports

Some dataflows might output data, while others may not.

For example, a dataflow may be designed to be called in response to an object being added to S3. The dataflow may fetch the data from S3,
parse the data, transform it, and then push the data into Apache Kafka. For such a dataflow, it probably doesn't make sense to output any data,
as the dataflow itself introduces a Kafka topic as the sink of the data.

Another dataflow, however, may be designed to be called in response to a message being placed onto an SNS topic. The dataflow may take that
notification, filter out messages that do not match some criteria, and then send to an Output Port any messages that do match the criteria.
That output may then be connected to another destination via an AWS Lambda Destination.

If no Output Port is present in the dataflow's root group, no output will be provided from the function invocation.

If one Output Port is present, and its name is "failure", any data routed to that Output Port will cause the function invocation to fail.
No output will be provided as the output of the function invocation.

If two Output Ports are present, and one of them is named "failure", that Output Port will behave as the "failure" port above.
The other Output Port is considered a successful invocation. When the dataflow is invoked, if a single FlowFile is routed to this
Port, the contents of the FlowFile will be emitted as the output of the function. If more than one FlowFile is sent to the Output Port,
the dataflow will be considered a failure, as Lambda requires that a single entity be provided as its output. A MergeContent or MergeRecord
Processor may be used in order to assemble multiple FlowFiles into a single output FlowFile if necessary.

Any Port with the name "failure" (ignoring case) will be considered a Failure Port. Additionally, if the "FAILURE_PORTS" environment variable
is specified, any port whose name is in the comma-separated list will be considered a Failure Port.

If two or more Output Ports, other than failure ports, are present, the "OUTPUT_PORT" environment variable must be provided. In such a case,
this environment variable must match the name of an Output Port at the dataflow's root group (case is NOT ignored). Anything that is routed
to the specified port is considered the output of the function invocation. Anything routed to any other port is considered a failure.



## Specifying Parameters

An important aspect of building a dataflow that is able to be run outside of the NiFi instance where the dataflow was built is the
concept of parameterization. NiFi allows users to build a dataflow making use of Parameter Contexts. This allows properties for
Processors and Controller Services to be defined at runtime instead of at build time.

Some parameters will be sensitive, such as a password, while others will not be, such as the name of a Kafka topic. The Lambda
function handler allows users to specify these parameters in one of two ways: through use of Environment Variables and through use
of the AWS Secrets Manager.  


### Environment Variables

Any Parameter can be specified using the environment variables of the AWS Lambda function. When configuring the Lambda function,
simply add an environment variable whose name matches the name of a Parameter in your Parameter Context.


### AWS Secrets Manager

A more secure mechanism for storing parameters is to use the AWS Secrets Manager. To do so, create a secret using the AWS Secrets Manager.
Select a secret type of "Other type of secrets (e.g. API key)." Enter one or more key/value pairs to use as Parameters. Each key provided
to the Secret Manager will map to a Parameter in the dataflow's Parameter Context with the same name.

If the name of the Secret matches the name of the Parameter Context in the dataflow, there is no further configuration needed. The Lambda
function will automatically configure the dataflow to pull the specified secret's values to represent the Parameter Context. However, there
are cases in which this does not work. Take, for example, a case in which a dataflow was developed with a Parameter Context named AWS_CONTEXT.
Now, consider that we want to deploy this dataflow three times. The first time, it should use the values in the "Context_1" Secret. For the second
deployment, it should use values for the "Context_2" Secret. And for the third deployment, it should use values from the "Context_3" Secret.

To accomplish this, we can specify an environment variable named PARAM_CONTEXT_AWS_CONTEXT. Here, the name of the environment variable is 
`PARAM_CONTEXT_` followed by the name of the Parameter Context. The value of the environment variable will be a comma-separated list of
Secret names from AWS Secrets Manager. So, for the first deployment we would use environment variable PARAM_CONTEXT_AWS_CONTEXT with a 
value of "Context_1".  For the second deployment, we would use an environment variable named PARAM_CONTEXT_AWS_CONTEXT with a value of 
"Context_2" and so on.

If our dataflow contains multiple Parameter Contexts, we can also map each of them to different secrets. For example, if we have a Parameter
Context named AWS_CONTEXT and one named CLOUDERA_CONTEXT, and we wanted to map those to secrets named "Context_1" and "Cldr_Context", respectively,
we could do so by adding two environment variables: PARAM_CONTEXT_AWS_CONTEXT = "Context_1" and CLOUDERA_CONTEXT = "Cldr_Context".

We can also specify that multiple Secrets map to the same Parameter Context by supplying a comma-separated list of Secret names.  For example,
if we want both the "Kafka" and "Common" Secrets to supply parameters to a Parameter Context named "ALL_KAFKA", we would set
PARAM_CONTEXT_ALL_KAFKA = "Kafka, Common".  For any key names in these Secrets that overlap, later Secrets override keys that are in
Secrets earlier in the list.

**Note:** Duplicate keys are unsupported, and will be non-deterministic.  Therefore, care should be taken that any Secrets that will be
mapped into the same Parameter Context do not contain the same key names, otherwise you won't be able to tell which Secret's key will be
used for the parameter value.

The same Secret may also supply parameters to multiple different Parameter Contexts.  Therefore,
a configuration like the following is possible:

```
PARAM_CONTEXT_KAFKA = "Kafka, Common"
PARAM_CONTEXT_SOLR = "Solr, Common"
```

In this case, the Common Secret would supply values to both the KAFKA and SOLR Parameter Contexts.

Additionally, we can simply default all Parameter Contexts whose names do not map Secrets in AWS Secrets Manager to use a default secret
by setting an environment variable named DEFAULT_PARAM_CONTEXT. The value of this environment variable should be the name of the Secret
to use.

It is important to note here that in order to use this capability, the Parameter Contexts used in the flow must have names that are compatible
with AWS Secrets Manager names and Environment Variable names. To ensure that these naming conventions are followed, Parameter Contexts and
Parameters should use names that start with a letter and consist only of letters, numbers, and the underscore (_) character. For example,
instead of naming a Parameter "AWS Parameter Context" use a name of AWS_PARAMETER_CONTEXT when building the dataflow.

It is also important to note that the user/role that is used to launch the Lambda function must have access to List Secrets in the AWS Secrets
Manager, as well as Read the secrets.


## Providing Custom Extensions / NARs

The binary that is deployed for running AWS Lambda does not include the full NiFi deployment with all NARs / extensions. Doing so would result
in very long startup times and would require more expensive configurations for Lambda. Instead, it packages only the NARs necessary to run Stateless
NiFi as well as the AWS NAR. If any other extension is needed in order to run the dataflow, it will be automatically downloaded when the Lambda
function is initialized on a Cold Start. The extensions will be downloaded from the Cloudera Maven Repository by default.

However, an alternate Nexus repository can be configured by setting the `NEXUS_URL` environment variable to the URL of the Nexus Repository where
extensions should be downloaded. For example, to use Google's mirror of the Maven Central repository, set the `NEXUS_URL` environment variable to
`https://maven-central.storage-download.googleapis.com/maven2`. Any configured URL must either be accessible via `http` or be served over `https` with
a trusted certificate.

Typically, this is the desired approach for downloading extensions that are made available by Cloudera or Apache NiFi. However, if there is a need
to provide custom extensions, a simpler route is usually to use Lambda's Layers capability. To do so, create a ZIP file containing all NAR files that
will need to be used. In the AWS console (or via the CLI), add a layer to your Lambda function, providing the ZIP file as the source for the layer.

Then, we simply need to tell Lambda where to look for those extensions. We do this by adding another environment variable whose name starts with
`EXTENSIONS_DIR_`. We can add multiple of these, such as `EXTENSIONS_DIR_1`, `EXTENSIONS_DIR_2`, `EXTENSIONS_DIR_MY_CUSTOM_LIBS`. The name of the
environment variable is the path where Lambda should look for the extensions. If the ZIP file provided as a Layer does not contain any directories,
the value should be `/opt`. If the ZIP file were to contain some directory, say `extensions/nars`, then the value for the environment variable would
be set to `/opt/extensions/nars`.

Adding a larger layer will result in the lambda configuration taking a bit longer to load on a Cold Start, and will take more memory to load the additional
classes. Therefore, it is not recommended to upload extremely large ZIP files containing hundreds of MBs or more of extensions if they are not needed.
However, this is supported if necessary. 

A third option for providing custom extensions is discussed below in the section on [S3 Bucket Storage](#s3-bucket-storage).

## Providing additional resources

In addition to NiFi extensions, some dataflows may also require additional resources in order to function. For example, a JDBC driver may be required for
establishing a database connection. Or a CSV file may be required in order to provide data enrichment. The recommended approach for providing such resources
to your Lambda function is via Lambda Layers.

Create a ZIP file containing any resources that are required for the dataflow to run. The ZIP file will then be unzipped into the `/opt` directory for access
by the dataflow.

Because each deployment of a given dataflow may need to reference files in a different location, depending on its environment, it is generally considered
a best practice to parameterize any resource that needs to be accessed by the dataflow. For example, if the dataflow contains a `DBCPConnectionPool` controller
service for accessing a database, it is recommended to use a Parameter for the "Database Driver Location(s)" property of the Controller Service. This allows each
deployment to specify the location of the JDBC Driver independently.

So, for example, we can set the "Database Driver Location(s)" property to a value of `#{JDBC_DRIVER_LOCATION}`. Then, in our Lambda, we can add a layer that
contains our JDBC driver, `database-jdbc-driver.jar`. And we can add an Environment Variable named `JDBC_DRIVER_LOCATION` with a value of `/opt/database-jdbc-driver.jar`.
By taking this approach, our dataflow is more reusable, as it can be deployed in many different environments.

A second option for providing dataflow resources is discussed below in the section on [S3 Bucket Storage](#s3-bucket-storage).

## S3 Bucket Storage

If extensions or resources of your dataflow exceed the size limit of Lambda Layers, you may provide these files in an S3 bucket.  
To do so, specify the S3 bucket name in the `STORAGE_BUCKET` environment variable.

Next, upload any extensions to a directory in the S3 bucket and specify the directory name in the `STORAGE_EXTENSIONS_DIRECTORY`
environment variable.  Upload any dataflow resources to another directory in the bucket and specify the directory name in
the `STORAGE_RESOURCES_DIRECTORY` environment variable.  Now when your Lambda Function does a cold start, it will download any extensions or
resources from these locations.

Note that in order to pull from S3, the Execution Role for the Lambda function must have both the ListBucket and GetObject permissions
for the chosen bucket.

# Configuring Kerberos

Depending on the dataflow that is used, you may need to enable Kerberos authentication in order to interact with some service(s). To enable Kerberos, we must provide
an appropriate `krb5.conf` file. This file tells NaaF, among other things, where the Key Distribution Center (KDC) is located and the Kerberos Realm to use.
To specify this, the `krb5.conf` must be made available to NaaF. This can be done, for instance, using the Lambda Layer or by placing the file in an S3 bucket.
See the section on [providing additional resources](#providing-additional-resources) to learn more.

Additionally, we must tell NiFi as a Function where it can find this `krb5.conf` file. To do this, add an environment variable to your Lambda function. The environment
variable must have the name `KRB5_FILE` and the value must be the (fully qualified) filename of the `krb5.conf` file. For example, it might use `/opt/krb5.conf`.

In order to use Keytabs in your dataflow, it will also be important to provide the necessary keytab files to the NaaF function. This is accomplished in the same manner,
using a Lambda Layer or an S3 bucket. While the `krb5.conf` file is a system configuration, the keytab is not. The keytab will be configured in the dataflow on a per-component
basis. For example, one Processor may use keytab `nifi-keytab-1` while some Controller Service makes use of keytab `hive-keytab-4`. The location of these files
may change from deployment to deployment. For example, for an on-premises deployment, keytabs may be stored in `/etc/security/keytabs` while in NaaF, they may be made available in `/opt`.
As a result, it is advisable to use Parameter Contexts in order to parameterize the location of the keytab files. This is a good practice for any file-based resource that will
need to be made available to a deployment. In this case, because there is a parameter referencing the location, we need to provide NaaF with a value for that parameter. To do
so, we can simply create an environment variable (see [specifying parameters](#specifying-parameters) for more information).

# Dataflow State

By default, if your dataflow contains any stateful processors (e.g., `ListSFTP`), this state will automatically be stored in a DynamoDB table called `nifi_state`,
though this name can be set using the `DYNAMODB_STATE_TABLE` environment variable.  If state is to be stored in DynamoDB, the role executing the Lambda function 
will require permissions to create and describe tables, and permissions to put and get items.  If the dataflow contains no stateful processors, this provider is 
not used, and will require no extra permissions.

The default DynamoDB billing mode is `PAY_PER_REQUEST`.  To change this to `PROVISIONED`, set the `DYNAMODB_BILLING_MODE` environment variable to
`PROVISIONED`, and set the `DYNAMODB_READ_CAPACITY_UNITS` and `DYNAMODB_WRITE_CAPACITY_UNITS` variables accordingly (see
https://docs.aws.amazon.com/amazondynamodb/latest/developerguide/HowItWorks.ReadWriteCapacityMode.html#HowItWorks.requests for more details).

The DynamoDB state provider can be disabled even if your dataflow contains stateful processors by setting the `DISABLE_STATE_PROVIDER` Environment Variable
to `true`.  Be warned, however, that this will cause processor state to be lost between function executions.

## Output Size Constraints

Often times, when we design a dataflow that will be deployed within AWS Lambda, we create a Process Group that contains a single Input Port (perhaps named `Input`)
and two Output Ports (perhaps named `success` and `failure`). If data is routed to the `failure` relationship, the entire invocation is failed. If data is routed to
the `success` port, the contents of the FlowFile become the output of the Lambda function. This allows us to set a Destination for our AWS Lambda. For example, we can
send the output to an SQS queue or another Lambda function.

However, we must keep in mind that AWS Lambda limits the size of the output. While the maximum output size is configurable, the default is 6 MB. So, by default, if
the size of the FlowFile routed to `success` exceeds 6 MB we end up with a failure - even if we have no Destination configured. It's also important to consider size
constraints on any downstream system that may be used as the Lambda Destination. Because of this, it is generally advisable not to send a large FlowFile as the output
of the Process Group. Rather, it may make sense to have the last Processor in the dataflow auto-terminate the `success` relationship so that there is no output FlowFile.
Or, it may make sense to use a `ReplaceText` processor with the "Replacement Strategy" set to `Always Replace`. This allows us to set the contents to some predefined text,
such as `{ "action": "perform dataflow", "success": true, "fileSize": ${fileSize} }`, for instance.


## Testing the Dataflow

As with any dataflow that we build, it is important to be able to test our dataflow before attempting to deploy it. Any dataflow that is to be used by
AWS Lambda will require an Input Port and most often will contain at least one Output Port. This makes it quite easy to test our dataflows in NiFi!
We can simply move up to the Parent Process Group within NiFi and create a GenerateFlowFile processor. The "Custom Text" property can then be set to any
value that we may wish to feed into our dataflow. For example, to simulate an event indicating that data was added to an S3 bucket, we might set the "Custom Text"
property to the value:

```
{
  "Records": [
    {
      "awsRegion": "us-east-1",
      "s3": {
        "bucket": {
          "name": "mpayne-nifi-logs",
          "arn": "arn:aws:s3:::example-bucket"
        },
        "object": {
          "key": "nifi-app_2021-10-12_18.36.log.gz",
          "size": 1024,
        }
      }
    }
  ]
}
```

We can then have this GenerateFlowFile processor connected to the Input Port of our Process Group that we will run in Lambda. And the "Run Once" feature of NiFi
makes it really easy to create this FlowFile and send it to the Process Group. We can then start our Process Group and ensure that the data processes as expected.
If not, we can fix our dataflow and trigger the GenerateFlowFile processor again, until we have properly handled the data.

We can then Download the Flow Definition from the NiFi and upload the flow to the Cloudera DataFlow Service catalog.


## Testing the Deployment

Once we've built and verified our dataflow and uploaded it to the Cloudera DataFlow Service catalog, we can create our Lambda. After creating the lambda
as described in the first section and configuring all necessary Environment Variables, we will want to test that we have configured all of the settings
before publishing the Lambda. To do so, we can use the "Test" tab of AWS Lambda. Lambda provides us with many options for sample data, so we can easily
create an example JSON document to simulate the events that our Lambda function will handle.

We can then click on the "Test" button to have the dataflow run. If there are problems, we can easily inspect the log messages to help us to understand the
problem. If the logs do not provide enough details, we may need to adjust the log levels in order to gain more debugging information. See the section below
for more details.

The amount of time the lambda takes to run depends heavily on the dataflow and how many extensions it needs. Because the Lambda function may have to download
some extensions from Nexus and perform initialization, a Cold Start may take many seconds. Approximately 20-30 seconds is not uncommon for a dataflow with
several extensions, while other dataflows may complete in 5 seconds. After we obtain a successful run of our function using the Test button, it may be helpful
to run several additional iterations in order to understand how the function will perform when a Cold Start is not necessary. Again, this depends heavily upon
the configured dataflow and other services that it may interact with. A simple dataflow may complete in as little as 10 milliseconds while a dataflow that must
perform complex transformations and reach out to one or more web services or databases for enrichment may take several seconds to complete.

It will be important to adjust Lambda's Timeout configuration in the "General configuration" tab of the Configuration screen if the Cold Start takes longer
than the amount of time allocated. 


## Adjusting Log Levels

NiFi as a Function makes use of the SLF4J Simple Logger for logging purposes. As such, we can update the log levels by adjusting the JVM's system properties.
To do so, we set the `JAVA_TOOL_OPTIONS` environment variable. To set a log level for a given logger, we set the JVM system property named 
`org.slf4j.simpleLogger.log.<logger name>` to the desired log level. For example, if we want to enable DEBUG logging for the `AwsSecretsManagerParameterValueProvider`
in order to understand whether or not Parameters are being fetched from AWS Secrets Manager, we would set the `JAVA_TOOL_OPTIONS` environment variable to
`-Dorg.slf4j.simpleLogger.log.org.apache.nifi.stateless.parameter.AwsSecretsManagerParameterValueProvider=DEBUG`.

We can also set additional logger levels by adding multiple `-D` options separated by spaces. For example, to enable DEBUG logs on both the
AWS Secrets Manager Parameter Value Provider and the Stateless Bootstrap class, we would set `JAVA_TOOL_OPTIONS` to a value of
`-Dorg.slf4j.simpleLogger.log.org.apache.nifi.stateless.parameter.AwsSecretsManagerParameterValueProvider=DEBUG -Dorg.slf4j.simpleLogger.log.org.apache.nifi.stateless.bootstrap.StatelessBootstrap=DEBUG`


## Handling Failures

As we mentioned, we can use a `failure` Output Port to notify Lambda that a Lambda invocation has failed. At that point, Lambda will automatically retry the invocation,
by default, twice. After that, if it still fails, the default option in Lambda is to simply drop the event notification. This could result in data loss, as it means
that our Lambda function will not have a chance to process the data again.

Instead, we may choose to use a "Dead-Leter Queue" for the data. To do this, in the AWS Console, we can go to Lambda's Configuration screen and choose the
"Asynchronous invocation" tab. Clicking the Edit button presents us with an option to use for the Dead-letter queue. We can use None (the default, which drops the notification),
or we can configure an SNS or SQS queue to use in the case of failure. This would give us the option to later go back and manually process the data or triage and
analyze it to understand why processing failed.
