# NaaF: Azure Function Apps

NiFi as a Function - Azure Function Apps

This module provides the necessary components for running NiFi as a Function (NiFi Stateless) as a Microsoft Azure's Function App.


## Creating Function App

To create a Function App  that is able to run NiFi as a Function, follow these steps.
1. Navigate to the Function App page (for example, https://portal.azure.com/#blade/HubsExtension/BrowseResource/resourceType/Microsoft.Web%2Fsites/kind/functionapp)
2. Click the "Create" button on the top-left of the page. 
3. Either select a Resource Group, or create a new one.  Enter a name in "Function App name", and leave "Code" selected next to "Publish".
   Select "Java" from "Runtime stack", and "11.0" from "Version".  Select the region appropriate to your use case.
4. You may create the Function App now by clicking the "Review + create" button, but there are other configuration options that
   can be set in this wizard which cannot be changed later.  We will review these by clicking the "Next : Hosting" button.
5. Here you see that a new Storage account will be created unless you specify an existing one.  You can also click the "Create new" link
   in order to name the new Storage account more descriptively.  Also notice that the Windows operating system has been selected.  You may
   choose Linux here if desired, but be warned that choosing Linux specifically on a "Consumption" plan is not currently recommended for NiFi as a Function
   due to a comparative lack of diagnostic, debugging, and logging support.
6. Also on this tab, you see the Plan type.  Azure has multiple plan types which should be weighed depending
   on your use case, but the default is "Consumption", meaning no dedicated servers are provisioned and your Function App is automatically
   scaled completely based on consumption.  There are many differences between these plans, including memory and timeout limits and 
   networking support, and a full comparison of the plans can be found here: https://docs.microsoft.com/en-us/azure/azure-functions/functions-scale
7. Click the "Next: Networking (preview)" button.  (Internal note for Cloudera testing: if you are testing a flow deployed in mow-priv or 
   mow-dev, this tab requires additional configuration.  Also note that this is not possible for Consumption plans, which will require your flow
   to be available in a public CDP instance.  If you are not using a Consumption plan, first ensure that your Region on the "Basics" tab 
   is set to in "West US 2".  Then in the Networking tab, select "On" for "Enable network injection", and select the "Virtual Network" 
   that matches your subscription name.  Leave "Inbound Access" Off, but set "Outbound Access" to "On".  Finally, select one of the subnets 
   named for the subscription (e.g., "eng-dim-longrunning.external.0.westus2").  At this point you can continue with the rest of the setup, 
   and your Function App will have access to the Cloudera internal network where mow-priv and mow-dev resources are deployed.)
8. Moving on, click "Next : Monitoring" and observe that Application Insights will automatically be integrated unless disabled here.
   At this point, you can safely click "Review + create", since the Tags can be set later.
9. Click the "Create" button to create your Function App, and wait for the deployment to complete.

At this point, some additional steps are required if you selected the App Service Plan Type:
1. If running an App Service plan, navigate to the storage account selected or created above, and click on the "Access keys" blade.  Click the "Show keys" button
   at the top, and copy the key1 "Connection string" value.
2. Back in the Function App, click on the Configuration blade, and the "New application setting" button to add a new setting named "WEBSITE_CONTENTAZUREFILECONNECTIONSTRING",
   and paste the copied value as the setting value.  Then add another Application setting named "WEBSITE_CONTENTSHARE" with a value of "<Your Function App Name>-fileshare".
3. Then click the Save button at the top, and Continue.  This will create a fileshare to back your Function App that you can easily access from your storage account.

If you selected a Consumption Linux plan from the Hosting tab, follow these steps to upload the NiFi as a Function binary:
1. Follow the instructions in the [Managed Identity](#managed-identity) section to create a managed identity for your Function App.
2. Navigate to the new storage account from above, and select the "Containers" blade, and then the "+ Container" button on the top left.  Specify a name for 
   a new Blob Container (e.g., "naaf"), and then click the "Create" button.
3. Click on the new container, and then click the "Access Control (IAM)" blade, and select the "Role assignments" tab.  Click the "Add" button
   at the top left, and specify "Add role assignment".  Then select "Storage Blob Data Reader" and click "Next".
4. Change "Assign access to" to "Managed identity", and then click the "Select members" link.  Under "Managed identity", select "Function App", and then click
   on the name of your Function App.  Click "Select", then "Review + assign".
5. Now go back to the new Container, and click the "Upload" button at the top left.  Upload the `NaafAzureFunctions.zip` file.
 
Otherwise, follow these steps for Windows or non-Consumption plans:
1. Navigate to the storage account selected or created above, and click on the "File shares" blade under "Data storage", and click
   the file share name.  Notice that this share has a 5TB quota.  This file share will be mapped to `/home` on the Function App filesystem.
2. Click on the "data" directory (create it if it does not yet exist), and then click the "Add directory" button at the top, to create a directory named "SitePackages".  Enter
    the new directory, and Upload the `NaafAzureFunctions.zip`.  Then, create a file named `packagename.txt` with the contents `NaafAzureFunctions.zip`,
    and upload it as well.

The Function app is now created and the binary is in place, but you'll need to configure the Application Settings in order to complete
the deployment.

## Configuration

### Configuring Application Settings

The function must now be configured to specify the dataflow to run and any necessary runtime configuration.
Configuration is provided using Environment Variables. Click the "Configuration" blade on the left, and observe the "Application Settings" section.
These settings are exposed as Environment Variables inside the function.  Azure pre-populates some of these, which are used by the Function
framework itself.  You will have to add the remaining Application Settings.

Click the "New application setting" button at the top to add the first Setting, which is required for all NiFi as a Function deployments.  
For the Name enter `WEBSITE_RUN_FROM_PACKAGE`.  The value is different depending on the Operating System:
- For Windows, enter `1`.  This tells the Function App to use the Zip package you uploaded earlier to the File share.
- For Linux, you'll need to enter the full path to the `NaafAzureFunctions.zip` uploaded to the Container above.  To get this URL,
  navigate to the Storage account, click "Containers", and then click on the container you created above.  Click on the `NaafAzureFunctions.zip` file, 
  and then click the Copy icon next to the URL.  Back in the function app, paste the value as the value of the`WEBSITE_RUN_FROM_PACKAGE` setting.

Additionally, the following Application Settings are supported.

| Variable Name | Description | Required | Default Value |
| ------------ | ----------- | -------- | ------------- |
| FUNCTION_NAME | The unique function name for this deployment.  Recommended to simply be the name of the Function App.  Warning: if more than one Function App has the same value for this variable, dataflow state may be corrupted. | true | -- |
| FLOW_CRN | The Cloudera Resource Name (CRN) for the dataflow that is to be run. The dataflow must be stored in the DataFlow Service's Catalog. This CRN should indicate the specific version of the dataflow and as such will end with some suffix such as `/v.1` | true | -- |
| DF_PRIVATE_KEY | The Private Key to use for accessing Cloudera's DataFlow service. The Private Key and Access Key are used to authenticate with the DataFlow Service and must provide the necessary authorizations to access the specified dataflow.  Note that it is a best practice to configure the `DF_PRIVATE_KEY` as a Secret in Key Vault.  See section below on [Key Vault Secret Parameters](#key_vault_secret_parameters) for more information. | true | -- |
| DF_ACCESS_KEY | The Access Key to use for accessing Cloudera's DataFlow service. The Private Key and Access Key are used to authenticate with the DataFlow Service and must provide the necessary authorizations to access the specified dataflow. | true | -- |
| INPUT_PORT | The name of the Input Port to use, if the dataflow has at least one.  If the dataflow has no Input Ports, this variable must not be specified. If the specified dataflow has more than one Input Port at the root group level, this environment variable must be specified, indicating the name of the Input Port to queue up the Cloud Function event. If only one Input Port exists, this variable is not required but if specified must properly match the name of the Input Port. | false | -- |
| OUTPUT_PORT | The name of the Output Port to retrieve the results from. If no Output Port exists, this does not need to be specified and no data will be returned. If at least one Output Port exists in the dataflow, this variable can be used to determine the name of the Output Port whose data will be sent along as output of the function. Otherwise, see the section below on [Output Ports](#output-ports) to understand how the appropriate Output Port is determined. | false | -- |
| FAILURE_PORTS | A comma-separated list of Output Ports that exist at the root group level of the dataflow. If any FlowFile is sent to one of these Output Ports, the function invocation is considered a failure. See section below on [Output Ports](#output-ports) for more information. | false | -- |
| DF_SERVICE_URL | The Base URL for the Cloudera Dataflow Service. | false | https://api.us-west-1.cdp.cloudera.com |
| NEXUS_URL | The Base URL for a Nexus Repository to use for downloading any NiFi Archives (NARs) needed for running the dataflow. | false | https://repository.cloudera.com/artifactory/cloudera-repos/ |
| CONTENT_REPO | The contents of the FlowFiles can be stored either in memory, on the JVM heap, or on disk. If this environment variable is specified, it specifies the path to a directory where the content should be stored. If not specified, the content is held in memory. | false | -- |
| WORKING_DIR | The working directory, where NAR files will be expanded.  See the section below on [Disk Usage Considerations](#disk-usage-considerations) for more information. | false | `/tmp/extensions` for Linux, `C:\local\Temp` for Windows |
| EXTENSIONS_DOWNLOAD_DIR | The directory to which missing extensions / NiFi Archives (NARs) will be downloaded. See the section below on Providing Custom Extensions / NARs for more information. | false | `/tmp/extensions` for Linux, `C:\local\Temp` for Windows |
| EXTENSIONS_DIR_* | Specifies read-only directories that may contain custom extensions / NiFi Archives (NARs). See the section below on Providing Custom Extensions / NARs for more information. | false | -- |
| STORAGE_ENDPOINT | A Storage account endpoint in which to look in for custom extensions / NiFi Archives (NARs) and resources. See the section below on [Storage Containers](#providing-extensions-and-resources-through-storage-containers) for more information. | false | -- |
| STORAGE_CONTAINER | A Storage account container name in which to look in for custom extensions / NiFi Archives (NARs) and resources. See the section below on [Storage Containers](#providing-extensions-and-resources-through-storage-containers) for more information. | false | -- |
| STORAGE_EXTENSIONS_DIRECTORY | The directory in the Storage account Blob container to look in for custom extensions / NiFi Archives (NARs). See the section below on [Storage Containers](#providing-extensions-and-resources-through-storage-containers) for more information. | false | extensions |
| STORAGE_RESOURCES_DIRECTORY | The directory in the Storage account Blob container look in for custom resources. See the section below on [Storage Containers](#providing-extensions-and-resources-through-storage-containers) for more information. | false | resources |
| COSMOSDB_ENDPOINT | The Cosmos DB account URI for the Cosmos DB state provider.  Required if you wish to preserve state in any stateful processors.  See [Dataflow State](#dataflow-state) for more information. | false | -- |
| COSMOSDB_STATE_DATABASE | The Cosmos DB database name for the Cosmos DB state provider.  See [Dataflow State](#dataflow-state) for more information. | false | `nifi_state` |
| COSMOSDB_STATE_CONTAINER | The Cosmos DB container name for the Cosmos DB state provider.  See [Dataflow State](#dataflow-state) for more information. | false | `nifi_state` |
| WEBSITE_RUN_FROM_PACKAGE | Specifies that the Function App should use the NaaF Zip package as its binary distribution. | true | 1 |
| KRB5_FILE | Specifies the filename of the krb5.conf file. This is necessary only if connecting to a Kerberos-protected endpoint. See section below on configuring Kerberos. | false | `/etc/krb5.conf` |
| REGION | Specifies the region of Azure account | false | -- |
| REGION | Specifies the region code of Azure account. For example, `westus` | false | -- |


Additionally, some Environment Variables apply only to HTTP triggers:

| Variable Name | Description | Required | Default Value |
| ------------- | ----------- | -------- | ------------- |
| HEADER_ATTRIBUTE_PATTERN | A RegEx capturing all flowfile attributes from the Output Port that should be returned as HTTP headers, if the trigger type is HTTP.  For all other trigger types, this variable is ignored. | false | -- |
| HTTP_STATUS_CODE_ATTRIBUTE | A flowfile attribute name that will set the response HTTP status code in a successful dataflow, if the trigger type is HTTP.  See the section below on [HTTP Triggers](#http-triggers) for more information.  For all other trigger types, this variable is ignored. | false | -- |

## Managed Identity

A Function App can use a Managed Identity in order to access other Azure services.

When configuring your Function App, click the "Identity" blade on the left.  Switch the "Status" to "On" and then click "Save", and then
finally "Yes", in order for Azure to create a Managed Identity to represent your Function App.  Your Function App now has an identity
in Azure Active Directory, and can be assigned roles and access to specific services like other identities.

When interacting with Key Vaults, Event Grid, or Storage accounts (aside from the File share automatically attached to the Function App), this
is the Identity you will use, and it is named after your Function App.  You can also grant it permissions to access other Azure services
if the dataflow itself requires access to them.

## Function Trigger Types

After you fully deploy your Function App, you can go the Functions blade to view the different Functions that come with NiFi as a Function.

1. `StatelessNiFiEventGridTriggerFunction` - Triggers based on an Azure EventGrid event.  This allows a wide range of possible events to 
   trigger your Function App.  See the section on [Event Grid Functions](#event-grid-functions) below for more information on configuring 
   Event Grid to interact with your Function App.
2. `StatelessNiFiHttpBinaryTriggerFunction` - Triggers in response to an HTTP request with a binary payload (`Content-Type: application/octet-stream`).
   See the section on [HTTP Functions](#http-functions) for more information this trigger type.
3. `StatelessNiFiHttpTextTriggerFunction` - Identical to `StatelessNiFiHttpBinaryTriggerFunction`, but supports text-based payloads such as
   JSON (`Content-Type: application/json`), XML (`Content-Type: application/xml`), etc.

Note that your dataflow will most likely be written differently between the two basic Function trigger types of Event Grid and HTTP, but
all three Functions are enabled by default.  To disable the other functions, click on the `...` button at the right of the Functions table
and then Disable.

### Event Grid Functions

The `StatelessNiFiEventGridTriggerFunction` can be used to trigger based on an Event Grid event, responding to a variety of Azure signals.

To set up an Event Grid Subscription for your Function App, follow these steps:
1. Navigate to the Azure "Event Grid System Topics" service, and click the "Create" button in the top-left.
2. Select the Topic Type accordingly, based on what service you want to trigger your Function App.  For example, if you want your Function
   to trigger every time a file is dropped into a Storage Container, select "Storage Accounts (Blob & GPv2)".  Select the Subscription of 
   the Resource you want to target, then select the "Resource Group" and "Resource" accordingly.
3. Enter a name for your System Topic, and check the "Enable system assigned identity" box.  Click the "Review + create" button, and then 
   "Create".  When the deployment completes, click "Go to resource".  Although you could have created an Event Grid Subscription directly, 
   creating a System Topic to house the Subscription allows additional diagnostics to be configured, such as routing information about
   DeliveryFailures to a Storage account, which is particularly useful in troubleshooting.  See the "Diagnostic settings" blade for more 
   information on configuring this setting.
4. Next, on the Event Grid System Topic "Overview" blade, click the "+ Event Subscription" button in the top-left.  Enter a name for the subscription.  Select the Event schema, which specifies
   the format of metadata sent as input to your NiFi dataflow.  The Event Grid schema can be found at https://docs.microsoft.com/en-us/azure/event-grid/event-schema,
   and the Cloud Event Schema v1.0 can be found at https://docs.microsoft.com/en-us/azure/event-grid/cloud-event-schema.
5. Optionally select the Event Types (for example, "Blob Created" for Storage accounts).  For Endpoint Details, select "Azure Function",
   and then click the "Select an endpoint" link.  Here your function will likely be automatically selected, but if not you can select
   the appropriate Function from the right panel, and then "Confirm Selection".  Finally, "Create".

You can now monitor the subscription from the System Topic, and should see any activity in the "Overview" blade.

### HTTP Functions

With HTTP functions, the Function is invoked by submitting an HTTP request, and the Function output from the dataflow is returned as
the HTTP response, with specific Output Port semantics discussed in the [HTTP Function Output Ports](#http-function-output-ports) 
section below.

Here is an example of invoking the `StatelessNiFiHttpTextTriggerFunction` using `curl`:

```
curl https://<function-app-name>.azurewebsites.net/api/StatelessNiFiHttpTextTriggerFunction \
     -H 'Content-Type: application/json' \
     --data-binary @my-file.json
```

The `StatelessNiFiHttpBinaryTriggerFunction` invocation is very similar, but allows binary data to be submitted:

```
curl https://<function-app-name>.azurewebsites.net/api/StatelessNiFiHttpBinaryTriggerFunction \
     -H 'Content-Type: application/octet-stream' \
     --data-binary @my-file.pdf
```

Note that the HTTP request body for Azure HTTP Function triggers is read completely into memory on the Heap, so it is recommended to limit
the size of the data sent in the HTTP request.  Microsoft's recommendation for working with large data objects is to store the objects
in a Storage account, and only pass around pointers to the objects in the HTTP request body.  NiFi dataflows can make use of this
recommendation by using the `FetchAzureBlobStorage` processor to retrieve the object if needed.

Also note that in any successful response, the `mime.type` flowfile attribute will be used as the `Content-Type` of the response.  Conversely,
the HTTP request `Content-Type` will be added as the `mime.type` attribute of the input flowfile.

There are two Application Settings that can be configured to customize the HTTP response:
- `HEADER_ATTRIBUTE_PATTERN` can provide a regular expression matching any flowfile attributes to be included as HTTP response headers.
  This means that any attributes to be included as headers should be named as the headers themselves.
- `HTTP_STATUS_CODE_ATTRIBUTE` specifies a flowfile attribute that sets the HTTP status code in the response.  This allows a code other
  than 200 for 'success' or 500 for 'failure' to be returned.  If this environment variable is specified but the attribute
  is not set, the status code in the HTTP response will default to 200.

It is important to note that in the "failure" scenario, the `HTTP_STATUS_CODE_ATTRIBUTE` is not used.  Therefore, if a non-500 status code
is required, it must be provided in a successful dataflow path.  For example, assume a `409 Conflict` is desired in a particular case.
This could be achieved as follows:
1. In the dataflow, include an `UpdateAttribute` processor that sets an arbitrary attribute, say `http.status.code`, to the value `409`.
2. Optionally, add a `ReplaceText` processor that sets the contents of the flowfile using `Always Replace` to whatever HTTP
   response body is desired.
3. Route this part of the flow to the main Output Port used for the success case.
4. In the GCP Cloud function, set the Runtime environment variable `HTTP_STATUS_CODE_ATTRIBUTE` to `http.status.code`, and Deploy.

This same method may also be used to set a non-200 success code.

Finally, all HTTP headers sent in the HTTP request will be imported as Flowfile attributes in the dataflow, with an attribute name like `azure.http.header.<lowercaseHeaderName>`.
For example, subbmitting a request with a header `X-My-Test-Header: test` will result in the flowfile having an attribute named `azure.http.header.x-my-test-header` with a value
of `test`.

## Providing Custom Extensions / NARs

The binary that is deployed for running NiFi as a Function does not include the full NiFi deployment with all NARs / extensions. Doing so would result
in very long startup times and would require more expensive configurations for the Function App. Instead, it packages only the NARs necessary to run 
Stateless NiFi. If any other extension is needed in order to run the dataflow, it will be automatically downloaded when the Function App is initialized 
on a Cold Start. The extensions will be downloaded from the Cloudera Maven Repository by default.

However, an alternate Nexus repository can be configured by setting the `NEXUS_URL` environment variable to the URL of the Nexus Repository where
extensions should be downloaded. For example, to use Google's mirror of the Maven Central repository, set the `NEXUS_URL` environment variable to
`https://maven-central.storage-download.googleapis.com/maven2`. Any configured URL must either be accessible via `http` or be served over `https` with
a trusted certificate.

Typically, this is the desired approach for downloading extensions that are made available by Cloudera or Apache NiFi. However, if there is a need
to provide custom extensions, a simpler route is usually to use the Storage account's File share.

### Providing Extensions from the File share

This process is different depending on the Plan Type and Operating System.  For every type except Linux on a Consumption plan, the File share 
was already created above, and you can simply create a directory in the file share containing all NAR files that are needed.

For Linux on a Consumption plan, you will need to create a File share and mount it as follows:
1. In the storage account attached to the Function App, click on the "File shares" blade, and then "+ File share" on the top left.  Name the File share, and then
   "Create".
2. Click the "Access keys" blade of the Storage account, then "Show keys".  Copy the Key from key1.  This will be pasted in place of \[Copied access key] below.
3. Click the "Cloud Shell" icon at the top-right, and accept any prompts needed to set up the shell.  When the shell is running, run the following command:
```
az webapp config storage-account add      \
   --resource-group [Resource Group Name] \
   --name [Function App Name]             \
   --custom-id [Pick a unique mount ID]   \
   --storage-type AzureFiles              \
   --share-name [File share Name]         \   
   --account-name [Storage account Name]  \  
   --mount-path /data                     \
   --access-key [Copied access key]
```
4. The mount path can be different, but should not be `/home` or `/tmp`.  Finally, you can create a directory in the File share, and upload any NAR files to the directory.

Then, regardless of OS or plan type, we simply need to tell the Function App where to look for those extensions. We do this by adding another Application 
Setting whose name starts with `EXTENSIONS_DIR_`. We can add multiple of these, such as `EXTENSIONS_DIR_1`, `EXTENSIONS_DIR_2`, 
`EXTENSIONS_DIR_MY_CUSTOM_LIBS`. The name of the Application Setting is the path where the Function App should look for the extensions.
- In Linux Consumption, the File share is mounted where you specified in the Cloud Shell above.  For example, if we upload our NARs to the File share under a directory
  called `extensions/custom`, and the File share is mounted under `/data`, we can add an Application Setting called `EXTENSIONS_DIR_CUSTOM` with a value of `/data/extensions/custom`.
- In all other plans, since the File share is mounted at `/home`, this should always start with `/home`.  For example, if we upload our NARs to the File share under a 
  directory called `extensions/custom`, we could add an Application Setting called `EXTENSIONS_DIR_CUSTOM` with a value of `/home/extensions/custom`.

Adding many extensions will result in the Function App configuration taking a bit longer to load on a Cold Start, and will take more memory to load the additional
classes. Therefore, it is not recommended to upload extremely large set of files containing hundreds of MBs or more of extensions if they are not needed.
However, this is supported if necessary.

There is a third option for providing custom extensions, which is discussed below in the section on
[Storage Containers](#providing-extensions-and-resources-through-storage-containers).

## Providing additional resources

In addition to NiFi extensions, some dataflows may also require additional resources in order to function. For example, a JDBC driver may be required for
establishing a database connection. Or a CSV file may be required in order to provide data enrichment. The recommended approach for providing such resources
to your Function App is via the attached Storage account's File share.  To set this up, follow the instructions in the 
[File share](#providing-extensions-from-the-file-share) section above.

Create a directory containing any resources that are required for the dataflow to run. The directory will then be mapped under the mapped File share directory
for access by the dataflow.

Because each deployment of a given dataflow may need to reference files in a different location, depending on its environment, it is generally considered
a best practice to parameterize any resource that needs to be accessed by the dataflow. For example, if the dataflow contains a `DBCPConnectionPool` controller
service for accessing a database, it is recommended to use a Parameter for the "Database Driver Location(s)" property of the Controller Service. This allows each
deployment to specify the location of the JDBC Driver independently.

So, for example, we can set the "Database Driver Location(s)" property to a value of `#{JDBC_DRIVER_LOCATION}`. Then, in our Storage File share, we can
create a directory called something like `resources`, which contains our JDBC driver `database-jdbc-driver.jar`. And we can add an Application Setting 
named `JDBC_DRIVER_LOCATION` with a value of `/home/resources/database-jdbc-driver.jar` (or `/data/resources/database-jdbc-driver.jar` for Linux Consumption plans).

By taking this approach, our dataflow is more reusable, as it can be deployed in many different environments.

There is another option for providing dataflow resources, which is discussed below in the section on
[Storage Containers](#providing-extensions-and-resources-through-storage-containers).

## Providing Extensions and Resources through Storage Containers

If extensions or resources are to be shared among many Function Apps, it can be convenient to have a single Container in a 
Storage account with one copy of the extensions or resources.  To integrate with an arbitrary Storage Container, specify the Storage 
account's endpoint in the `STORAGE_ENDPOINT` Application Setting (e.g., `https://<storageAcount>.blob.core.windows.net`), and the 
Container name in the `STORAGE_CONTAINER` Application Setting.

Your Function App will need permission to access the container.  Assuming you have created a [Managed Identity](#managed-identity) for
your Function App, edit your Function App and click the "Identity" blade on the left.  Click the "Azure role assignments" button under
"Permissions", and then "Add role assignment".  Under "Scope", select "Storage", and then select the desired Storage account under
"Resource".  "Storage Blob Data Reader" should be sufficient to download Blobs.  Click "Save".

Next, upload any extensions to a directory in the Container and specify the directory name in the `STORAGE_EXTENSIONS_DIRECTORY`
Application Setting.  Upload any dataflow resources to another directory in the Container and specify the directory name in
the `STORAGE_RESOURCES_DIRECTORY` Application Setting.  Now when your App Function is restarted, it will download any extensions or 
resources from these locations.

Note: if you are using the Storage Container approach, you may see the following errors in the log at function startup, which can be safely ignored:

```
2022-01-21T18:36:33.976 [Error] [pool-2-thread-4] ERROR com.azure.identity.EnvironmentCredential - 
Azure Identity => ERROR in EnvironmentCredential: Missing required environment variable AZURE_CLIENT_ID
2022-01-21T18:36:34.460 [Error] [pool-2-thread-4] ERROR com.azure.identity.implementation.IntelliJCacheAccessor - IntelliJ Authentication not available. Please log in with Azure Tools for IntelliJ plugin in the IDE.
2022-01-21T18:36:34.917 [Error] [parallel-2] ERROR com.azure.identity.EnvironmentCredential - EnvironmentCredential authentication unavailable. Environment variables are not fully configured.To mitigate this issue, 
please refer to the troubleshooting guidelines here at https://aka.ms/azsdk/net/identity/environmentcredential/troubleshoot
2022-01-21T18:36:35.779 [Error] [pool-2-thread-4] ERROR com.azure.identity.EnvironmentCredential - Azure Identity => ERROR in EnvironmentCredential: Missing required environment variable AZURE_CLIENT_ID
2022-01-21T18:36:36.142 [Error] [pool-2-thread-4] ERROR com.azure.identity.implementation.IntelliJCacheAccessor - IntelliJ Authentication not available. Please log in with Azure Tools for IntelliJ plugin in the IDE.
2022-01-21T18:36:36.451 [Error] [parallel-2] ERROR com.azure.identity.EnvironmentCredential - EnvironmentCredential authentication unavailable. Environment variables are not fully configured.To mitigate this issue, 
please refer to the troubleshooting guidelines here at https://aka.ms/azsdk/net/identity/environmentcredential/troubleshoot
```

## Disk Usage Considerations
Azure Function Apps have two basic options for local storage, such as is used for downloading and expanding NARs:

1. Temporary storage, under `C:\local\Temp` - for Consumption plans, the **total** local disk size is limited to 500 MB, including this
   Temp directory, making the actual limit somewhat less than 500 MB.  This limit can be increased significantly with App Service and Functions Premium
   plans (for example, the Basic plan starts at 11 GB).  This storage is attached directly to the Function App container, and is therefore much
   faster than the File Share option, discussed next.  This storage also requires no additional configuration, but caution must be taken not
   to fill up the disk.  Mitigations for this are discussed below.
2. File Share storage, which is mapped to `C:\home`.  When setting up the initial function, you already attached a File share from your Storage account,
   so this should already be available to you.  The File share has a default limit of 5 TB, but is not locally attached to your Function App container,
   and so can be significantly slower than writing to the Temp directory.

Although Temporary storage is faster, NiFi as a Function has a storage optimization that reuses downloaded and unpacked extension NARs.  Therefore,
the recommended approach is to use the File share for both the `WORKING_DIR` and `EXTENSIONS_DOWNLOAD_DIR`.  This will incur a significant time penalty
on the very first cold start, but afterwards the downloading and unpacking of NARs can be skipped, greatly reducing the function execution time.

Note: Because of the initial execution time cost, when using the File share for either directory, it is recommended that you perform a single "warming"
invocation of your function in order to download and unpack all extension NARs for later reuse.

You can use the File share for storage by doing the following:

1. Navigate to your Storage account and click on the `File shares` blade on the left.  Select your file share name on the right.
2. Click on the `data` directory (or create it if it does not yet exist).
3. Click the `+ Add directory` button and name it `extensions`, and click `OK`.  Repeat this for a directory called `working`.
4. In your Function App, click the `Configuration` blade on the left, and add (or set) an Application Setting named `EXTENSIONS_DOWNLOAD_DIR`
   with a value of `C:\home\data\extensions`.
5. In your Function App, click the `Configuration` blade on the left, and add (or set) an Application Setting named `WORKING_DIR` with a value of
   `C:\home\data\working`.
6. Save the configuration.
7. Optionally (but recommended) perform a single "warming" invocation of your Function App.


# Configuring Kerberos

Depending on the dataflow that is used, you may need to enable Kerberos authentication in order to interact with some service(s). To enable Kerberos, we must provide
an appropriate `krb5.conf` file. This file tells NaaF, among other things, where the Key Distribution Center (KDC) is located and the Kerberos Realm to use.
To specify this, the `krb5.conf` must be made available to NaaF. This can be done, for instance, using an Azure Storage Container.
See the section on [providing additional resources](#providing-additional-resources) to learn more.

Additionally, we must tell NiFi as a Function where it can find this `krb5.conf` file. To do this, add an environment variable to your function. The environment
variable must have the name `KRB5_FILE` and the value must be the (fully qualified) filename of the `krb5.conf` file. For example, `/home/resources/krb5.conf`.

In order to use Keytabs in your dataflow, it will also be important to provide the necessary keytab files to the NaaF function. This is accomplished in the same manner,
using an Azure Storage Container. While the `krb5.conf` file is a system configuration, the keytab is not. The keytab will be configured in the dataflow on a per-component
basis. For example, one Processor may use keytab `nifi-keytab-1` while some Controller Service makes use of keytab `hive-keytab-4`. The location of these files
may change from deployment to deployment. For example, for an on-premises deployment, keytabs may be stored in `/etc/security/keytabs` while in NaaF, they may be made available in `/home/resources`.
As a result, it is advisable to use Parameter Contexts in order to parameterize the location of the keytab files. This is a good practice for any file-based resource that will
need to be made available to a deployment. In this case, because there is a parameter referencing the location, we need to provide NaaF with a value for that parameter. To do
so, we can simply create an environment variable (see [specifying parameters](#specifying-parameters) for more information).



# Dataflow State

By default, if your dataflow contains any stateful processors (e.g., `ListSFTP`), this state will automatically be stored in a Cosmos database called `nifi_state`,
though this name can be set using the `COSMOSDB_STATE_DATABASE` environment variable.  In order to integrate with Cosmos DB, some additional steps are required.
Without these steps, the dataflow will still run, but the state will not be preserved.

1. In the Azure console, navigate to the Cosmos DB service.
2. Click the "Create" button.  Click "Create" under "Core (SQL) - Recommended".
3. Select your Resource Group, and enter an account name.
4. All other configuration is optional, and may be different depending on your use case.  While provisioned throughput is supported, it is recommended
   to select "Serverless". Click "Review + Create", and then "Create".
5. In your Function App, go to the Identity blade, and note the Object (principal) ID, for the following steps.
6. Click on the Cloud Shell icon in the upper right.  Follow any initial prompts that may set up the web-based cloud shell for you.
7. Run the following commands in the cloud shell:
```
resourceGroup=<your resource group>
accountName=<your CosmosDB account name>
principalId=<your Function App Object principal ID from above>
az cosmosdb sql role assignment create --account-name $accountName --resource-group $resourceGroup --scope "/" \
  --principal-id $principalId --role-definition-id 00000000-0000-0000-0000-000000000001
az cosmosdb sql role assignment create --account-name $accountName --resource-group $resourceGroup --scope "/" \
  --principal-id $principalId --role-definition-id 00000000-0000-0000-0000-000000000002
```
8. On the "Overview" blade, click the "Add Container" button.  Specify the database and container IDs (if you use `nifi_state` for both, the
    `COSMOSDB_STATE_DATABASE` and `COSMOSDB_STATE_CONTAINER` Application Settings are not needed).  Specify `/id` for the Partition Key. Click "OK".
9. In your Function App Configuration, add an Application Setting called `COSMOSDB_ENDPOINT` with the Cosmos DB URI (from the Overview page).
   If you specified any other database or container name than `nifi_state`, you will need to specify the `COSMOSDB_STATE_DATABASE` and 
   `COSMOSDB_STATE_CONTAINER` Application Settings, respectively.

The Cosmos DB state provider can be disabled even if your dataflow contains stateful processors by setting the `DISABLE_STATE_PROVIDER` Application Setting
to `true`.  Be warned, however, that this will cause processor state to be lost between function executions.

## Output Ports

While all Function trigger types support Failure ports, only HTTP Functions will use the output from non-failure Output Ports.
Here we examine the different considerations of Output Ports for the different Function trigger types.

### Event Grid Function Output Ports

With Event Grid Functions, the Output Port dataflow logic is simpler.

If one Output Port is present, and its name is "failure", any data routed to that Output Port will cause the function invocation to fail.

If two Output Ports are present, and one of them is named "failure", that Output Port will behave as the "failure" port above.
Any other non-failure Output Port is considered a successful invocation, but the function will do nothing with data routed here.

Any Port with the name "failure" (ignoring case) will be considered a Failure Port. Additionally, if the "FAILURE_PORTS" environment variable
is specified, any port whose name is in the comma-separated list will be considered a Failure Port.

### HTTP Function Output Ports

Since Functions with an HTTP trigger are invoked synchronously, they directly return a response.  The Output Port semantics
for this case is as follows.

If no Output Port is present in the dataflow's root group, no output will be provided from the function invocation, but a default
success message will be returned in the HTTP response with a status code of 200.

If one Output Port is present, and its name is "failure", any data routed to that Output Port will cause the function invocation to fail.
No output will be provided as the output of the function invocation, and an error message will be returned from the HTTP response,
with a status code of 500.

If two Output Ports are present, and one of them is named "failure", that Output Port will behave as the "failure" port above.
The other Output Port is considered a successful invocation. When the dataflow is invoked, if a single FlowFile is routed to this
Port, the contents of the FlowFile will be returned as the HTTP response. If more than one FlowFile is sent to the Output Port,
the dataflow will be considered a failure, as there is only one HTTP response possible. A MergeContent or MergeRecord
Processor may be used in order to assemble multiple FlowFiles into a single output FlowFile if necessary.

Any Port with the name "failure" (ignoring case) will be considered a Failure Port. Additionally, if the "FAILURE_PORTS" environment variable
is specified, any port whose name is in the comma-separated list will be considered a Failure Port.

If two or more Output Ports, other than failure ports, are present, the "OUTPUT_PORT" environment variable must be provided. In such a case,
this environment variable must match the name of an Output Port at the dataflow's root group (case is NOT ignored). Anything that is routed
to the specified port is considered the output of the function invocation. Anything routed to any other port is considered a failure.

Also note that in any successful response, the `mime.type` flowfile attribute will be used as the `Content-Type` of the response.  

## Specifying Parameters

An important aspect of building a dataflow that is able to be run outside of the NiFi instance where the dataflow was built is the
concept of parameterization. NiFi allows users to build a dataflow making use of Parameter Contexts. This allows properties for
Processors and Controller Services to be defined at runtime instead of at build time.

Some parameters will be sensitive, such as a password, while others will not be, such as the name of a Kafka topic. The Cloud
function allows users to specify these parameters through use of Application Settings, either directly or mapped from Secrets.

### Application Settings Parameters

Any Parameter can be specified using the ApplicationSettings of the Azure Function App. When configuring the Function App,
in the "Configuration" blade, simply add an Application Setting whose name matches the name of a Parameter in your Parameter Context.


### Key Vault Secret Parameters

A more secure mechanism for storing sensitive dataflow parameters and other sensitive Application Settings is to use a Key Vault Secret.
To do so, navigate to the "Key Vaults" Azure service and click the "Create" button at the top-left.  Select your Function App's resource
group, and name the Key Vault.  Select the same region as the Function App.

Click the "Next : Access policy" button, and "Add Access Policy".  Under "Secret permissions", select "List" and "Get".  Then click the
"None selected" link next to "Select principal".  Assuming you have created a [Managed Identity](#managed-identity) for
your Function App, type the name of your Function App and select it on the right panel, then click
the "Select" button.  Click the "Add" button, then the "Review + create" button, and finally "Create".

Once the deployment is complete, click the "Go to resource" button.  Here you can create any secrets to be referenced, as follows:
1. Click the "Secrets" blade on the left, and then the "Generate/Import" button at the top.
2. In "Name", enter the name of the desired Application Setting.  This does not have to match the actual name of the Application Setting
   or Parameter Context, and must follow the convention of only alphanumeric characters and dashes.
3. In the "Value", enter the value of the Application Setting, and click Create.

Back in your Function App, create an Application Setting with the appropriate name (the Parameter name, or one of the supported
Application Settings from the [Configuring Application Settings](#configuring-application-settings) section).  In the Value,
enter the following:
```
@Microsoft.KeyVault(VaultName={vaultName};SecretName={secretName})`
```

Replace `{vaultName}` with the actual Key Vault name, and replace `{secretName}` with the name of the newly created secret.

When you click the Save button, Azure will update your configuration and then attempt to link this Application Setting to the Key Vault
Secret.  To see this link, you can click the Refresh button.  In the Source column, you should now see "Key vault Reference", and
a green check mark if the Function App successfully resolved the Secret.  If there is a red `X`, there is likely an issue with either
the permissions of the Function App's managed identity or the vault or secret could not be found.

It is important to note here that in order to use this capability for dataflow parameters, the Parameter Contexts used in the flow must 
have names that are compatible with Azure Application Settings names. To ensure that these naming conventions are followed, Parameter Contexts and
Parameters should use names that start with a letter and consist only of letters, numbers, and the underscore (_) character. For example,
instead of naming a Parameter "Azure Parameter Context" use a name of AZURE_PARAMETER_CONTEXT when building the dataflow.

## Testing the Dataflow

As with any dataflow that we build, it is important to be able to test our dataflow before attempting to deploy it. Any dataflow that is to be used by
Azure Function Apps will require an Input Port and will most often contain at least one failure Output Port. This makes it quite easy to test our dataflows in NiFi!
We can simply move up to the Parent Process Group within NiFi and create a GenerateFlowFile processor. The "Custom Text" property can then be set to any
value that we may wish to feed into our dataflow. For example, to simulate an Event Grid event indicating that data was added to a Storage Container, we might set the "Custom Text"
property to the value:

```
{
  "topic": "/subscriptions/73db9d72-863b-45e9-9a1a-890cff0ac0e6/resourceGroups/my-rg/providers/Microsoft.Storage/storageAccounts/mystorageaccount",
  "subject": "/blobServices/default/containers/my-container/blobs/input/testfile.json",
  "eventType": "Microsoft.Storage.BlobCreated",
  "id": "d4c93b9a-901e-0078-1dd4-d34b8d06bbad",
  "data": {
    "api": "DeleteBlob",
    "requestId": "9d4c93b9-901e-0078-1dd4-d34b8d000000",
    "eTag": "0xD9A1EBFB6C05A58",
    "contentType": "application/json",
    "contentLength": 136,
    "blobType": "BlockBlob",
    "url": "https://mystorageaccount.blob.core.windows.net/my-container/input/testfile.json",
    "sequencer": "000000000000000000000000000004F50000000003e3cc87",
    "storageDiagnostics": {
      "batchId": "69a23459-7006-003d-00d4-d39e6e000000"
    }
  },
  "dataVersion": "",
  "metadataVersion": "1",
  "eventTime": "2021-11-07T12:41:55.7182885Z"
}
```

We can then have this GenerateFlowFile processor connected to the Input Port of our Process Group that we will run in the Function App. And the "Run Once" feature of NiFi
makes it really easy to create this FlowFile and send it to the Process Group. We can then start our Process Group and ensure that the data processes as expected.
If not, we can fix our dataflow and trigger the GenerateFlowFile processor again, until we have properly handled the data.

We can then Download the Flow Definition from the NiFi and upload the flow to the Cloudera DataFlow Service catalog.


## Testing the Deployment

Once we've built and verified our dataflow and uploaded it to the Cloudera DataFlow Service catalog, we can create our Function App. After deploying the function
as described in the first section, we will want to test that we have configured all of the settings. To do so, we can click on the "Functions" blade of our Function App,
and then click on the Function we want to test.

We can then click on the "Code + Test" blade on the left, and then the "Test/Run" button at the top.  This brings up a side panel to input sample data and have the dataflow run.
If there are problems, we can easily inspect the log messages to help us to understand the issue. If the logs do not provide enough details, we may need to adjust the log levels 
in order to gain more debugging information, and view the complete Log Stream as described in the [Logging](#logging) section below.

Note: at this time, the Azure Console cannot submit valid requests in the "Code + Test" view specifically for Event Grid trigger functions.  To work around this, you can test your Function
without generating an actual Event Grid event as follows:

1. Click the "Get function URL" button at the top of the "Code + Test" view, and click the Copy button next to the URL.
2. Create a file, for example, named `event-grid.json` with the contents of the Event Grid schema JSON you want to test.
3. Paste the URL in place of the URL in the following `curl` command:
```
curl https://<app-function-name>.azurewebsites.net/runtime/webhooks/EventGrid?functionName=StatelessNiFiEventGridTriggerFunction&code=<code> \
     -H 'Content-Type: application/json'\
     -H 'aeg-event-type: Notification'
     --data-binary @event-grid.json
```

See https://docs.microsoft.com/en-us/azure/azure-functions/functions-bindings-event-grid-trigger?tabs=java%2Cbash#manually-post-the-request for more details on submitting this request.

The amount of time the cloud function takes to run depends heavily on the dataflow and how many extensions it needs. Because the Cloud function may have to download
some extensions from Nexus and perform initialization, a Cold Start may take many seconds. Approximately 20-30 seconds is not uncommon for a dataflow with
several extensions, while other dataflows may complete in 10 seconds. After we obtain a successful run of our function using the Test button, it may be helpful
to run several additional iterations in order to understand how the function will perform when a Cold Start is not necessary. Again, this depends heavily upon
the configured dataflow and other services that it may interact with. A simple dataflow may complete in as little as 5 milliseconds for HTTP triggers or 100 milliseconds for Event Grid triggers,
while a dataflow that must perform complex transformations and reach out to one or more web services or databases for enrichment may take several seconds to complete.

It will be important to adjust Cloud function's Timeout configuration in the "Runtime" tab of the "Runtime, build, connections and security settings" if the Cold Start takes longer
than the amount of time allocated. 

# Securing your App

The following steps are recommended in order to secure the HTTP endpoint for your App.

## Enforce HTTPS

In your Function App, click on the "TLS/SSL Settings" blade on the left.  Next to "HTTPS Only", click "On".  Also ensure that "Minimum TLS Version" is the
latest version.  This page is also where you may configure the server certificate if desired.

## Add an Identity Provider

By default, your HTTP Function endpoint is completely public.  In order to restrict access to only authorized users, you can add an Identity Provider:

1. In the Authentication blade of your Function App, click the "Add Identity Provider" button.  Several options are available, but in this example, we will select Microsoft.
2. Leave the App registration type as "Create new app registration".  Enter the name of your Function App as the registration "Name".
3. Select the desired "Supported account types" (for example, "Any Azure AD directory - Multi-tenant")
4. Ensure "Restrict access" is set to "Require authentication".
5. Select the desired "Unauthenticated requests" value.  For this example, we will select "HTTP 401 Unauthorized: recommended for APIs".
6. Ensure "Token store" is selected.
7. Click the "Add" button.

The identity provider will now appear in the Authentication blade.  You can follow the "Quickstart" link for details on how to use it to authenticate for your
HTTP Function, but we will continue the example with a common use case here.

To access your HTTP Function endpoint programmatically, you can create a new App Registration representing the calling client:

1. In the App Registrations service, click the "New registration" button.  Give the registration a name representing your client (e.g., myfunction-client).
2. Select the desired "Supported account types" option, and then "Register".
3. In the "Certificates & secrets" blade, select "Client secrets", and then "New client secret".  Specify a description, and click "Add".  Copy the secret Value,
   because it will never be shown again.
4. In the Overview blade, you will need the "Directory (tenant) ID" and "Application (client) ID" for the next step.  You will also need the "Application (client) ID"
   of your Function App's App Registration for the next step.
5. Request an access token similar to the following:

```
# <tenant-id> is the tenant ID copied above
# <client-id> is the client App Registration's "Application (client) ID" copied above
# <client-secret> is the client App Registration's client secret copied above
# <application-id> is your Function App's App Registration "Application (client) ID" copied above
curl -X POST https://login.microsoftonline.com/<tenant-id>/oauth2/token -H "Content-Type: application/x-www-form-urlencoded" -d \
    'grant_type=client_credentials&client_id=<client-id>&client_secret=<client-secret>&resource=<application-id>' -X POST
```

6. This will give you an OAuth 2 response.  Copy the value of the "access_token" to use below.
7. Submit a request to your HTTP function now as follows:

```
TOKEN=<access_token> # copied from above
curl https://<your-function-app>.azurewebsites.net/api/StatelessNiFiHttpTextTriggerFunction \
    -H "Authorization: Bearer $TOKEN" \
    ... etc.
```

## Logging

NiFi as a Function makes use of the SLF4J Simple Logger for logging purposes. As such, we can update the log levels by adjusting the JVM's system properties.
To do so, we set the `JAVA_TOOL_OPTIONS` Application Setting. To set a log level for a given logger, we set the JVM system property named 
`org.slf4j.simpleLogger.log.<logger name>` to the desired log level. For example, if we want to enable DEBUG logging for the `StandardStatelessNiFiFunction`
in order to view the actual data flow JSON being used, we would set the `JAVA_TOOL_OPTIONS` environment variable to
`-Dorg.slf4j.simpleLogger.log.org.com.cloudera.naaf.StandardStatelessNiFiFunction=DEBUG`.

We can also set additional logger levels by adding multiple `-D` options separated by spaces. For example, to enable DEBUG logs on both the
NiFi as a Function framework and the Stateless Bootstrap class, we would set `JAVA_TOOL_OPTIONS` to a value of
`-Dorg.slf4j.simpleLogger.log.com.cloudera.naaf.StandardStatelessNiFiFunction=DEBUG -Dorg.slf4j.simpleLogger.log.org.apache.nifi.stateless.bootstrap.StatelessBootstrap=DEBUG`

Function logs can be viewed either directly in the "Log Stream" blade at the Function App level, or under the "Monitoring" blade of the individual Function.  Function output log files
are also found in the Storage account File share under `LogFiles/Application/Functions/Host`.

## Handling Failures

As we mentioned, we can use a `failure` Output Port to notify the Function that an invocation has failed. For HTTP triggers, there is no retry.  For Event Grid triggers, Azure Event Grid may 
automatically retry the invocation, if configured in the Event Subscription, for a default of 30 attempts. After that, if it still fails, the Event Subscription will simply drop the 
event notification. This could result in data loss, as it means that our Function App will not have a chance to process the data again.
