/*
 * (c) 2018-2021 Cloudera, Inc. All rights reserved.
 *
 *   This code is provided to you pursuant to your written agreement with Cloudera, which may be the terms of the
 *  Affero General Public License version 3 (AGPLv3), or pursuant to a written agreement with a third party authorized
 *  to distribute this code.  If you do not have a written agreement with Cloudera or with an authorized and
 *  properly licensed third party, you do not have any rights to this code.
 *
 *   If this code is provided to you under the terms of the AGPLv3:
 *   (A) CLOUDERA PROVIDES THIS CODE TO YOU WITHOUT WARRANTIES OF ANY KIND;
 *   (B) CLOUDERA DISCLAIMS ANY AND ALL EXPRESS AND IMPLIED WARRANTIES WITH RESPECT TO THIS CODE, INCLUDING BUT NOT
 *       LIMITED TO IMPLIED WARRANTIES OF TITLE, NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE;
 *   (C) CLOUDERA IS NOT LIABLE TO YOU, AND WILL NOT DEFEND, INDEMNIFY, OR HOLD YOU HARMLESS FOR ANY CLAIMS ARISING
 *       FROM OR RELATED TO THE CODE; AND
 *   (D) WITH RESPECT TO YOUR EXERCISE OF ANY RIGHTS GRANTED TO YOU FOR THE CODE, CLOUDERA IS NOT LIABLE FOR ANY
 *       DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR CONSEQUENTIAL DAMAGES INCLUDING, BUT NOT LIMITED
 *       TO, DAMAGES RELATED TO LOST REVENUE, LOST PROFITS, LOSS OF INCOME, LOSS OF BUSINESS ADVANTAGE OR
 *       UNAVAILABILITY, OR LOSS OR CORRUPTION OF DATA.
 *
 */

package com.cloudera.naaf;

import com.cloudera.naaf.metering.CloudPlatform;

import java.io.File;
import java.util.Map;
import java.util.Optional;
import java.util.function.Consumer;

public class StatelessFunctionConfiguration {

    private static final File DEFAULT_EXTENSION_DIRECTORY = new File(StatelessNiFiFunction.TEMP_DIR, "extensions");

    private final Consumer<Map<String, String>> flowDefinitionConfigAction;
    private final String region;
    private final CloudPlatform cloudPlatform;
    private final ClassLoader systemClassLoader;
    private final String resourceDownloaderType;
    private final String stateProviderType;
    private final File resourcesDirectory;
    private final String resourcesSourceDirectory;
    private final String extensionsSourceDirectory;
    private final File defaultExtensionDirectory;
    private final File narDirectory;

    private StatelessFunctionConfiguration(final Builder builder) {
        this.flowDefinitionConfigAction = builder.flowDefinitionConfigAction;
        this.defaultExtensionDirectory = builder.defaultExtensionDirectory;
        this.resourcesDirectory = builder.resourcesDirectory;
        this.resourcesSourceDirectory = builder.resourcesSourceDirectory;
        this.extensionsSourceDirectory = builder.extensionsSourceDirectory;
        this.systemClassLoader = builder.systemClassLoader;
        this.resourceDownloaderType = builder.resourceDownloaderType;
        this.stateProviderType = builder.stateProviderType;
        this.narDirectory = builder.narDirectory;
        this.region = builder.region;
        this.cloudPlatform = builder.cloudPlatform;
    }

    public void applyFlowDefinitionConfigAction(final Map<String, String> flowDefinition) {
        Optional.ofNullable(flowDefinitionConfigAction).ifPresent(action -> action.accept(flowDefinition));
    }

    public File getDefaultExtensionDirectory() {
        return Optional.ofNullable(defaultExtensionDirectory).orElse(DEFAULT_EXTENSION_DIRECTORY);
    }

    public ClassLoader getSystemClassLoader() {
        return systemClassLoader;
    }

    public File getResourcesDirectory() {
        return resourcesDirectory;
    }

    public File getNarDirectory() {
        return narDirectory;
    }

    public String getResourceDownloaderType() {
        return resourceDownloaderType;
    }

    public String getStateProviderType() {
        return stateProviderType;
    }

    public String getResourcesSourceDirectory() {
        return resourcesSourceDirectory;
    }

    public String getExtensionsSourceDirectory() {
        return extensionsSourceDirectory;
    }

    public CloudPlatform getCloudPlatform() {
        return cloudPlatform;
    }

    public String getRegion() {
        return region;
    }

    public static class Builder {
        private Consumer<Map<String, String>> flowDefinitionConfigAction;
        private String region;
        private File resourcesDirectory;
        private String resourcesSourceDirectory;
        private String extensionsSourceDirectory;
        private File narDirectory;
        private File defaultExtensionDirectory;
        private String resourceDownloaderType;
        private String stateProviderType;
        private ClassLoader systemClassLoader;
        private CloudPlatform cloudPlatform;

        public Builder flowDefinitionConfigAction(final Consumer<Map<String, String>> action) {
            this.flowDefinitionConfigAction = action;
            return this;
        }

        public Builder region(final String region) {
            this.region = region;
            return this;
        }

        public Builder cloudPlatform(final CloudPlatform cloudPlatform) {
            this.cloudPlatform = cloudPlatform;
            return this;
        }

        public Builder systemClassLoader(final ClassLoader systemClassLoader) {
            this.systemClassLoader = systemClassLoader;
            return this;
        }

        public Builder resourceDownloaderType(final String resourceDownloaderType) {
            this.resourceDownloaderType = resourceDownloaderType;
            return this;
        }

        public Builder stateProviderType(final String stateProviderType) {
            this.stateProviderType = stateProviderType;
            return this;
        }

        public Builder resourcesDirectory(final File resourcesDirectory) {
            this.resourcesDirectory = resourcesDirectory;
            return this;
        }

        public Builder narDirectory(final File narDirectory) {
            this.narDirectory = narDirectory;
            return this;
        }

        public Builder resourcesSourceDirectory(final String resourcesSourceDirectory) {
            this.resourcesSourceDirectory = resourcesSourceDirectory;
            return this;
        }

        public Builder extensionsSourceDirectory(final String extensionsSourceDirectory) {
            this.extensionsSourceDirectory = extensionsSourceDirectory;
            return this;
        }

        public Builder defaultExtensionDirectory(final File defaultExtensionDirectory) {
            this.defaultExtensionDirectory = defaultExtensionDirectory;
            return this;
        }

        public StatelessFunctionConfiguration build() {
            return new StatelessFunctionConfiguration(this);
        }
    }
}
